# -*- coding: utf-8 -*-


from __future__ import unicode_literals

import re

from . import errors
from ._compat import unicode
from .utils import lstripped, peekable
from .tokens import (
    TOKENS, NewLine, Indent, Dedent, End,
    LeftBrace, LeftSquare, LeftRound, RightBrace, RightSquare, RightRound,
)


#: Flags to use in the Scanner class.
SCANNER_FLAGS = re.MULTILINE | re.UNICODE | re.VERBOSE


def _tokenize(input_string):
    position = len(lstripped(input_string)) + 1
    tokens, remainder = _scanner.scan(input_string.strip())
    tokens = peekable(tokens)

    curr_indent = 0
    indent_stack = [0]
    newline_last = False
    inside_bracket = 0

    while tokens:
        indent_change = 0
        tok = tokens.next()
        tok.line = position

        # If inside brackets, no Indent/Dedent tokens are yielded.
        if tok.id in (LeftRound.id, LeftSquare.id, LeftBrace.id):
            inside_bracket += 1
        elif tok.id in (RightRound.id, RightSquare.id, RightBrace.id):
            inside_bracket -= 1

        # Determination of current indentation and indentation change
        # is necessary for correct generation of the Indent/Dedent tokens.
        if newline_last and not inside_bracket:
            indent = tok.value if tok.id == Indent.id else 0
            if indent != curr_indent:
                indent_change = indent - curr_indent
                curr_indent = indent

        # Here we determine the position of a token in the input string.
        if tok.id == NewLine.id:
            newline_last = True
            while tokens.peek().id == NewLine.id:
                tok.value = tokens.next().value
            position += tok.value
        else:
            newline_last = False

        # If indentation decreased we want to generate the needed dedent
        # tokens. These tokens are instantiated here as they cannot be
        # matched by regular expression.
        if indent_change < 0:
            while indent_stack[-1] > curr_indent:
                yield Dedent(indent_stack.pop(), line=position)
                yield NewLine(1, line=position)

        # If indentation increased we want to yield the indent token.
        if indent_change > 0:
            indent_stack.append(curr_indent)
            yield tok

        # We don't want to yield any other Indent tokens as our goal is
        # to represent the left/right braces with Indent/Dedent tokens.
        if tok.id != Indent.id and \
           not (inside_bracket and tok.id == NewLine.id):
            yield tok

    while len(indent_stack) > 1:
        yield Dedent(indent_stack.pop(), line=position)


class tokenize(peekable):
    """Tokenizes an input string.

    :param input_string: String to be tokenized.
    :type input_string: str
    :return: List of pairs (token type, value).
    """
    _marker = End()

    def __init__(self, input_string):
        tokens = _tokenize(unicode(input_string))
        super(tokenize, self).__init__(tokens)

    def advance(self, allowed=None, skip=None):
        """Helper for iterating through tokens.

        :param allowed: Optional list of allowed tokens. Default is
            any token. If the found token is not allowed, the function
            raises syntax  error.
        :type allowed: :class:`Token` or iterable of tokens
        :param skip: If :obj:`True`, a sequence of given token types
            is skipped first. Default is :obj:`False`.
        :type skip: boolean
        """
        tok = self.next()
        if skip is not None:
            while tok.id == skip.id:
                tok = self.next()
        if allowed is None:
            return tok
        if not isinstance(allowed, (list, tuple)):
            allowed = [allowed]
        if all(tok.id != Token.id for Token in allowed):
            raise_error(allowed, tok)
        return tok


def raise_error(expected, token):
    """Raises an error with some information about position etc.

    :param expected: List of expected tokens.
    :param token: Received token.
    :raises: :class:`errors.ParserError`
    """
    msg = 'Unexpected {}'.format(token.name)
    if token.line:
        msg += ' on line {}'.format(token.line)
    if expected and token.id != Indent.id:
        allowed_list = [Token.name for Token in expected if Token.re]
        if allowed_list:
            tok_msg = ' or '.join(allowed_list)
            msg += ', expected {}'.format(tok_msg)
    raise errors.ParserError(msg + '.')


def parse(input_string):
    """Parses given string according to NEON syntax.

    :param input_string: String to parse.
    :type input_string: string
    :return: Parsed string.
    :rtype: :class:`OrderedDict`
    """
    tokens = tokenize(input_string)
    return Indent().parse(tokens)


#: The Scanner is instantiated with a list of re's and associated
#: functions. It is used to scan a string, returning a list of parts
#: which match the given re's.
#:
#: See: http://stackoverflow.com/a/17214398/2874089
_scanner = re.Scanner([
    TokenClass.getscan() for TokenClass in TOKENS
    if TokenClass.re is not None
], flags=SCANNER_FLAGS)



# -*- coding: utf-8 -*-


from __future__ import unicode_literals


def format_list(list_, indent_level):
    indent = '\t' * indent_level
    newline = '\n' if indent_level else '\n\n'
    return newline.join([
        '{}- {}'.format(indent, to_string(value, indent_level + 1))
        for value in list_
    ])


def format_dict(dict_, indent_level):
    indent = '\t' * indent_level
    newline = '\n' if indent_level else '\n\n'
    return newline.join([
        '{}{}: {}'.format(indent, key, to_string(value, indent_level + 1))
        for key, value in dict_.items()
    ])


def to_string(obj, indent_level=0):
    """Encodes given object using the NEON syntax.

    :param obj: Object to encode.
    :return: Encoded object.
    :rtype: string
    """
    if isinstance(obj, dict):
        return '\n' + format_dict(obj, indent_level)
    elif isinstance(obj, list):
        return '\n' + format_list(obj, indent_level)
    elif obj is None:
        return 'Null'
    return str(obj)



# -*- coding: utf-8 -*-


from __future__ import unicode_literals

import re
import itertools

from ._compat import unicode


class classproperty(object):
    """Useful when class properties need to be defined."""

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, cls=None):
        if cls is None:
            cls = type(obj)
        return self.fget(cls)


def lstripped(string):
    """Number of potentially stripped characters on left.

    :return: A string of characters that will be stripped
        on the left side of the given string if the method
        `strip()` is called on it.
    :rtype: string
    """
    return ''.join(itertools.takewhile(unicode.isspace, string))


def variants(*strings):
    """Creates three variants of each string:

    - lowercase (e.g. `husky`)
    - title version (e.g. `Husky`)
    - uppercase (e.g. `HUSKY`)

    :return: A list of all variants of all given strings.
    :rtype: list
    """
    result = []
    for string in strings:
        lowercase = string.lower()
        result += [lowercase, lowercase.title(), string.upper()]
    return result


def camel_case_to_underscore(name):
    """Converts string from camel case notation to underscore.

    :param name: String to convert to underscore.
    :type name: string
    :return: A string converted from camel case to underscore.
    :rtype: string
    """
    s1 = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class peekable(object):
    """Wrapper for an iterator to allow 1-item lookahead
    Call ``peek()`` on the result to get the value that will next pop out of
    ``next()``, without advancing the iterator:
        >>> p = peekable(xrange(2))
        >>> p.peek()
        0
        >>> p.next()
        0
        >>> p.peek()
        1
        >>> p.next()
        1
    Pass ``peek()`` a default value, and it will be returned in the case where
    the iterator is exhausted:
        >>> p = peekable([])
        >>> p.peek('hi')
        'hi'
    If no default is provided, ``peek()`` raises ``StopIteration`` when there
    are no items left.
    To test whether there are more items in the iterator, examine the
    peekable's truth value. If it is truthy, there are more items.
        >>> assert peekable(xrange(1))
        >>> assert not peekable([])

    .. NOTE:: Taken from: https://github.com/erikrose/more-itertools
    """
    # Lowercase to blend in with itertools. The fact that it's a class is an
    # implementation detail.

    _marker = object()

    def __init__(self, iterable):
        self._it = iter(iterable)

    def __iter__(self):
        return self

    def __nonzero__(self):
        try:
            self.peek()
        except StopIteration:
            return False
        return True

    def peek(self, default=_marker):
        """Return the item that will be next returned from ``next()``.
        Return ``default`` if there are no items left. If ``default`` is not
        provided, raise ``StopIteration``.
        """
        if not hasattr(self, '_peek'):
            try:
                self._peek = next(self._it)
            except StopIteration:
                if default is self._marker:
                    raise
                self._peek = self._marker
        return self._peek

    def next(self):
        ret = self.peek()
        del self._peek
        return ret



# -*- coding: utf-8 -*-


from collections import OrderedDict


class Entity(object):
    """Representation of Foo(bar=1) literal."""

    def __init__(self, value=None, attrs=None):
        self.value = value
        self.attributes = OrderedDict(attrs) or OrderedDict()

    def __repr__(self):
        keywords = ', '.join([
            '{}={}'.format(key, value) if pos != key else str(value)
            for pos, (key, value) in enumerate(self.attributes.items())
        ])
        return '{}({})'.format(self.value, keywords)

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        return (
            self.value == other.value
            and
            self.attributes == other.attributes
        )



# -*- coding: utf-8 -*-


from __future__ import unicode_literals

import re
import dateutil.parser
from collections import OrderedDict

from . import errors
from .entity import Entity
from .utils import variants, classproperty, camel_case_to_underscore


#: List of all tokens.
TOKENS = []

#: Pattern for matching hexadecimal numbers.
PATTERN_HEX = re.compile(r'0x[0-9a-fA-F]+')


def token(cls):
    """Registers a token class.
    """
    assert issubclass(cls, Token), 'Tokens must subclass the Token class.'
    TOKENS.append(cls)
    return cls


class Token(object):
    """Token representation.
    """
    #: Regular expression for tokenization.
    re = None

    @classproperty
    def id(cls):
        return cls

    @classproperty
    def name(cls):
        return camel_case_to_underscore(cls.__name__).replace('_', ' ')

    def __init__(self, value=None, line=None):
        self.value = value
        self.line = line

    def parse(self, tokens):
        return self.value

    def __eq__(self, other):
        return type(self) == type(other) and self.value == other.value

    def __str__(self):
        name = type(self).__name__
        value = '' if self.value is None else self.value
        return '{}({})'.format(name, value)

    def __repr__(self):
        return str(self)

    @classmethod
    def do(cls, scanner, string):
        return cls(string)

    @classmethod
    def getscan(cls):
        return (cls.re, cls.do)


class Primitive(Token):
    """Represents primitive type.
    """
    def parse(self, tokens):
        peek = tokens.peek()
        if peek.id == LeftRound.id:
            attributes = tokens.advance().parse(tokens)
            return Entity(self.value, attributes)
        return self.value


@token
class String(Primitive):
    """Represents string token.
    """
    re = r"""
          (?: "[^"\n]*" | '[^'\n]*' )
          """

    @classmethod
    def do(cls, scanner, string):
        double = '"'
        single = "'"
        if string[0] == double:
            string = string.strip(double)
        else:
            string = string.strip(single)
        return cls(string)


@token
class Integer(Primitive):
    """Represents integer token.
    """
    re = None

    @classmethod
    def convert(cls, string):
        try:
            if string.isdigit():
                return int(string)
            if PATTERN_HEX.match(string):
                return int(string, base=16)
        except ValueError:
            return


@token
class Float(Primitive):
    """Represents float token.
    """
    re = None

    @classmethod
    def convert(cls, string):
        try:
            return float(string)
        except ValueError:
            return


@token
class Boolean(Primitive):
    """Represents boolean token.
    """
    re = None

    _mapping = {
        True: variants('true', 'yes', 'on'),
        False: variants('false', 'no', 'off'),
    }

    @classmethod
    def convert(cls, string):
        for value, alternatives in cls._mapping.items():
            if string in alternatives:
                return value


@token
class NoneValue(Primitive):
    """Represents :obj:`None` token.
    """
    re = None

    _variants = variants('null')


@token
class DateTime(Primitive):
    """Represents datetime token.
    """
    re = None

    @classmethod
    def convert(cls, string):
        try:
            return dateutil.parser.parse(string)
        except ValueError:
            return


@token
class Literal(Token):
    """Represents literal token.
    """
    re = r"""
          (?: [^#"',:=[\]{}()\x00-\x20!`-] | [:-][^"',\]})\s] )
          (?: [^,:=\]})(\x00-\x20]+ | :(?! [\s,\]})] | $ ) |
              [\ \t]+ [^#,:=\]})(\x00-\x20] )*
          """

    @classmethod
    def do(cls, scanner, string):
        for Type in [Integer, Float, Boolean, DateTime]:
            value = Type.convert(string)
            if value is not None:
                return Type(value)
        if string in NoneValue._variants:
            return NoneValue(None)
        return String(string)


class Symbol(Token):
    """Represents symbol token.
    """
    @classproperty
    def name(cls):
        return "'{}'".format(str(cls.re).replace('\\', ''))

    @classmethod
    def do(cls, scanner, string):
        return cls()


@token
class Comma(Symbol):
    """Represents comma token.
    """
    re = r','


@token
class Colon(Symbol):
    """Represents colon token.
    """
    re = r':'


@token
class EqualSign(Symbol):
    """Represents equal sign.
    """
    re = r'='


@token
class Hyphen(Symbol):
    """Represents hyphen token.
    """
    re = r'-'


@token
class LeftRound(Symbol):
    """Represents left round bracket.
    """
    re = r'\('

    def parse(self, tokens):
        data = OrderedDict()
        tok = tokens.advance(skip=NewLine)
        iteration = 0

        while tok.id != RightRound.id:
            key = tok.parse(tokens)
            tok = tokens.advance((EqualSign, Comma, RightRound))

            if tok.id == EqualSign.id:
                data[key] = tokens.advance().parse(tokens)
                tok = tokens.advance((Comma, RightRound))
                if tok.id == Comma.id:
                    tok = tokens.advance(skip=NewLine)

            elif tok.id == Comma.id:
                data[iteration] = key
                tok = tokens.advance(skip=NewLine)

            elif tok.id == RightRound.id:
                data[iteration] = key

            iteration += 1

        return data


@token
class RightRound(Symbol):
    """Represents right round bracket.
    """
    re = r'\)'


@token
class LeftSquare(Symbol):
    """Represents left square bracket.
    """
    re = r'\['

    def parse(self, tokens):
        data = []
        tok = tokens.advance(skip=NewLine)

        while tok.id != RightSquare.id:
            value = tok.parse(tokens)
            data.append(value)

            tok = tokens.advance((Comma, RightSquare))
            if tok.id == Comma.id:
                tok = tokens.advance(skip=NewLine)

        return data


@token
class RightSquare(Symbol):
    """Represents right square bracket.
    """
    re = r'\]'


@token
class LeftBrace(Symbol):
    """Represents left brace.
    """
    re = r'{'

    def parse(self, tokens):
        data = OrderedDict()
        tok = tokens.advance(skip=NewLine)

        while tok.id != RightBrace.id:
            key = tok.parse(tokens)
            tokens.advance(Colon)
            data[key] = tokens.advance().parse(tokens)

            tok = tokens.advance((Comma, RightBrace))
            if tok.id == Comma.id:
                tok = tokens.advance(skip=NewLine)

        return data


@token
class RightBrace(Symbol):
    """Represents right brace.
    """
    re = r'}'


@token
class Comment(Token):
    """Represents comment token.
    """
    re = r'\#.*'
    do = None  # ignore comments


@token
class Indent(Token):
    """Represents indent token.
    """
    re = r'^[\t\ ]+'

    def _parse_list(self, tokens):
        data = []
        tok = tokens.advance()

        while tok.id not in [Dedent.id, End.id]:
            if tokens.peek().id == NewLine.id:
                value = None
            else:
                tok = tokens.advance()
                if tokens.peek().id == Colon.id:
                    tokens.advance()
                    key = tok.parse(tokens)
                    tok = tokens.advance(skip=NewLine)
                    value = {key: tok.parse(tokens)}
                else:
                    value = tok.parse(tokens)
            data.append(value)

            tok = tokens.advance((End, NewLine, Dedent))
            if tok.id == NewLine.id:
                tok = tokens.advance((Hyphen, Dedent))

        return data

    def _parse_dict(self, tokens, tok=None):
        data = OrderedDict()
        tok = tok or tokens.advance()

        while tok.id not in [Dedent.id, End.id]:
            key = tok.parse(tokens)
            tokens.advance(Colon)

            tok = tokens.advance()
            if tok.id == NewLine.id:
                tok = tokens.advance()
                if tok.id not in [Indent.id, Dedent.id]:
                    data[key] = None
                    continue
            data[key] = tok.parse(tokens)

            tok = tokens.advance((End, NewLine, Dedent))
            if tok.id == NewLine.id:
                tok = tokens.advance(skip=NewLine)

        return data

    def parse(self, tokens):
        peek = tokens.peek()

        while peek.id == NewLine.id:
            tokens.advance()
            peek = tokens.peek()

        if peek.id == Hyphen.id:
            return self._parse_list(tokens)
        else:
            return self._parse_dict(tokens)

    @classmethod
    def do(cls, scanner, string):
        return cls(len(string))


@token
class Dedent(Token):
    """Represents dedent token.
    """
    re = None  # this token is generated after the scanning procedure


@token
class NewLine(Token):
    """Represents new line token.
    """
    re = r'[\n]+'

    @classmethod
    def do(cls, scanner, string):
        return cls(len(string))


@token
class WhiteSpace(Token):
    """Represents comment token.
    """
    re = r'[\t\ ]+'
    do = None  # ignore white-spaces


@token
class Unknown(Token):
    """Represents unknown character sequence match.
    """
    re = r'.*'

    @classmethod
    def do(cls, scanner, token):
        msg = 'Unknown character sequence: {!r}'
        raise errors.TokenError(msg.format(token))


@token
class End(Token):
    """Represents EOL token.
    """
    re = None
    name = 'end of file'



# -*- coding: utf-8 -*-


__author__ = 'Pavel Dedik'
__version__ = '0.1.1'


from .decoder import parse
from .encoder import to_string


__all__ = ('decode', 'encode')


def decode(config):
    try:
        string = config.read()
    except AttributeError:
        string = config
    return parse(string)


def encode(tree):
    return to_string(tree)



# -*- coding: utf-8 -*-


import re
from setuptools import setup, find_packages


# determine version and the author
code = open('neon/__init__.py', 'r').read(1000)
version = re.search(r'__version__ = \'([^\']*)\'', code).group(1)
author = re.search(r'__author__ = \'([^\']*)\'', code).group(1)


setup(
    name='neon-py',
    version=version,
    author=author,
    description='NEON parser for Python',
    long_description=open('README.rst').read(),
    keywords=['neon', 'parser', 'config file'],
    author_email='dedikx@gmail.com',
    url='https://github.com/paveldedik/neon-py',
    license=open('LICENSE').read(),
    packages=find_packages(exclude=['tests']),
    install_requires=['python-dateutil==2.5.3'],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Topic :: Utilities',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ]
)
